/*
 * NoScript Commons Library
 * Reusable building blocks for cross-browser security/privacy WebExtensions.
 * Copyright (C) 2020-2024 Giorgio Maone <https://maone.net>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

// depends on nscl/content/patchWindow.js
// depends on nscl/content/patchWorkers.js

"use strict";
ns.on("capabilities", event => {

  if (!ns.canScript || ns.allows("webgl") ||
      !("HTMLCanvasElement" in window && document.createElement("canvas").getContext("webgl"))) {

    return;
  }

  function modifyGetContext(scope, {port, xray}) {
    let dispatchEvent = EventTarget.prototype.dispatchEvent;
    let { Event } = scope;
    for (let canvas of ["HTMLCanvasElement", "OffscreenCanvas"]) {
      if (!(canvas in scope)) continue;

      // CAVEAT:
      // we must use the X-Ray wrapper from window/globalThis for instanceof,
      // but proxy the wrapped getContext method from unprivileged scope, see
      // https://forums.informaction.com/viewtopic.php?p=104382

      const CanvasClass = globalThis[canvas];
      // globalThis future-proofs us for when we dare patchWorkers()

      const getContext = xray.getSafeMethod(scope[canvas].prototype, "getContext");

      const MAX_CONSECUTIVE = 20;
      let consecutive = 0;
      let lastTime = 0;
      let panic = false;
      const handler = cloneInto({
        apply: function(targetObj, thisArg, argumentsList) {
          debug(`WebGLHook called from ${new Error().stack}, ${thisArg}, ${canvas}, ${canvas?.parentElement}`);
          if (thisArg instanceof CanvasClass && /webgl/i.test(argumentsList[0])) {
            if (panic) {
              return null;
            }
            const target = canvas === "HTMLCanvasElement" && document.contains(thisArg) ? thisArg : scope;
            const t = Date.now();
            if (t - lastTime < 5 && consecutive++ > MAX_CONSECUTIVE) {
              console.error("Too many consecutive blocked webgl contexts, trying to break the loop.");
              panic = true;
              port.postMessage("panic");
            } else {
              port.postMessage("notify", target);
              lastTime = t;
            }
            return null;
          }
          return getContext.call(thisArg, ...argumentsList);
        }
      }, scope, {cloneFunctions: true});

      const proxy = new scope.Proxy(getContext, handler);
      scope[canvas].prototype.getContext = proxy;
    }
  }

  const notifyWebGL = canvas => {
    let request = {
      id: "noscript-webgl",
      type: "webgl",
      url: document.URL,
      documentUrl: document.URL,
      embeddingDocument: true,
    };
    seen.record({policyType: "webgl", request, allowed: false});
    notifyPage();
    if (canvas && !(canvas instanceof HTMLCanvasElement)) {
      request.offscreen = true;
      canvas = null;
    }
    try {
      let ph = PlaceHolder.create("webgl", request);
      ph.replace(canvas);
      PlaceHolder.listen();
    } catch (e) {
      error(e);
    }
  }

  function panicAbort() {
    const html = document.documentElement.outerHTML;
    const scriptBlocker = `<head><meta http-equiv="content-security-policy" content="script-src 'none'"></head>`;
    DocRewriter.rewrite(scriptBlocker);
    DocRewriter.rewrite(html);

    const target = document.body.appendChild(document.createElement("canvas"));
    target.style = "position: fixed; top: 0; left: 0; right: 0; bottom: 0; width: 100%; height: 100%";
    notifyWebGL(target);
  }

  let port = patchWindow(modifyGetContext);
  port.onMessage = (msg, {target}) => {
    debug(`WebGLHook msg: ${msg}, target: ${target}`);
    switch(msg) {
      case "notify":
        notifyWebGL(target);
        break;
      case "panic":
        panicAbort();
        break;
    }
  }

  debug(`WebGLHook installed on window ${document.URL}.`);

  if (!(globalThis.OffscreenCanvas && new OffscreenCanvas(0,0).getContext("webgl"))) {

    return;
  }

  try {
    const channelID = `webglHook:${uuid()}`;
    const bc = new BroadcastChannel(channelID);
    bc.onmessage = notifyWebGL;
    const workersPatch = () => {

      const bc = new BroadcastChannel(channelID);
      const getContext = OffscreenCanvas.prototype.getContext;
      const handler = {
        apply: function(targetObj, thisArg, argumentsList) {

          if (/webgl/i.test(argumentsList[0])) {
            bc.postMessage({});
            return null;
          }
          return getContext.call(thisArg, ...argumentsList);
        }
      };
      OffscreenCanvas.prototype.getContext = new Proxy(getContext, handler);
    };
    patchWorkers(`(${workersPatch})()`.replace(/\bchannelID\b/g, JSON.stringify(channelID)));

  } catch(e) {
    error(e);
  }
});
